/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2020 Realtek Technologies Co., Ltd
 */

#ifndef _R8125_HW_H_
#define _R8125_HW_H_

#include <stdint.h>

#include <rte_ethdev.h>
#include <rte_ethdev_core.h>

#include "r8125_base.h"
#include "r8125_ethdev.h"

u16 rtl8125_mac_ocp_read(struct rtl_hw *hw, u16 addr);
void rtl8125_mac_ocp_write(struct rtl_hw *hw, u16 addr, u16 value);

void ClearAndSetPCIePhyBit(struct rtl_hw *hw, u8 addr, u16 clearmask, u16 setmask);
void ClearPCIePhyBit(struct rtl_hw *hw, u8 addr, u16 mask);
void SetPCIePhyBit( struct rtl_hw *hw, u8 addr, u16 mask);

u16 rtl8125_ephy_read(struct rtl_hw *hw, int addr);
void rtl8125_ephy_write(struct rtl_hw *hw, int addr, int value);

u32 rtl8125_mdio_read(struct rtl_hw *hw, u32 addr);
void rtl8125_mdio_write(struct rtl_hw *hw, u32 addr, u32 value);

u32 rtl8125_eri_read(struct rtl_hw *hw, int addr, int len, int type);
int rtl8125_eri_write(struct rtl_hw *hw, int addr, int len, u32 value, int type);

u32 rtl8125_csi_read(struct rtl_hw *hw, u32 addr);

void rtl8125_get_mac_version(struct rtl_hw *hw);
int rtl8125_get_mac_address(struct rtl_hw *hw, struct rte_ether_addr *ea);

void rtl8125_enable_cfg9346_write(struct rtl_hw *hw);
void rtl8125_disable_cfg9346_write(struct rtl_hw *hw);

void rtl8125_rar_set(struct rtl_hw *hw, uint8_t *addr);

void rtl8125_wait_txrx_fifo_empty(struct rtl_hw *hw);

void rtl8125_disable_now_is_oob(struct rtl_hw *hw);
void rtl8125_init_ll_share_fifo(struct rtl_hw *hw);

void rtl8125_hw_disable_mac_mcu_bps(struct rtl_hw *hw);

void rtl8125_disable_aspm_clkreq_internal(struct rtl_hw *hw);
void rtl8125_disable_ups(struct rtl_hw *hw);

void rtl8125_enable_rxdvgate(struct rtl_hw *hw);
void rtl8125_disable_rxdvgate(struct rtl_hw *hw);

void rtl8125_init_software_variable(struct rtl_hw *hw);
void rtl8125_link_option(u8 *aut, u32 *spd, u8 *dup, u32 *adv);

void rtl8125_get_tally_stats(struct rtl_hw *hw, struct rte_eth_stats *stats);
void rtl8125_clear_tally_stats(struct rtl_hw *hw);

void rtl8125_exit_oob(struct rtl_hw *hw);
void rtl8125_nic_reset(struct rtl_hw *hw);

void rtl8125_set_hw_ops(struct rtl_hw *hw);

void rtl8125_switch_mac_mcu_ram_code_page(struct rtl_hw *hw, u16 page);
void _rtl8125_write_mac_mcu_ram_code(struct rtl_hw *hw, const u16 *entry, u16 entry_cnt);
void _rtl8125_write_mac_mcu_ram_code_with_page(struct rtl_hw *hw, const u16 *entry, u16 entry_cnt, u16 page_size);
void rtl8125_write_mac_mcu_ram_code(struct rtl_hw *hw, const u16 *entry, u16 entry_cnt);

extern const struct rtl_hw_ops rtl8125a_ops;
extern const struct rtl_hw_ops rtl8125b_ops;

#define RTL8125_MAC_MCU_PAGE_SIZE 256 //256 words

#define ARRAY_SIZE(arr) sizeof(arr) / sizeof((arr)[0])
#define HW_HAS_WRITE_PHY_MCU_RAM_CODE(_M)        (((_M)->HwHasWrRamCodeToMicroP == TRUE) ? 1 : 0)
#define HW_SUPPORT_MAC_MCU(_M)        ((_M)->HwSuppMacMcuVer > 0)
#endif
