/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2020 Realtek Technologies Co., Ltd */

#include <sys/queue.h>
#include <stdio.h>
#include <errno.h>
#include <stdint.h>
#include <stdarg.h>

#include <rte_ether.h>
#include <ethdev_driver.h>

#include "r8125_ethdev.h"
#include "r8125_hw.h"
#include "r8125_phy.h"

#define NO_BASE_ADDRESS 0x0
//extern int rtl8125_phy_ram_code_check(struct rtl_hw *hw);//tocheck

u16 map_phy_ocp_addr(u16 PageNum, u8 RegNum)
{
        u16 OcpPageNum = 0;
        u8 OcpRegNum = 0;
        u16 OcpPhyAddress = 0;

        if ( PageNum == 0 ) {
                OcpPageNum = OCP_STD_PHY_BASE_PAGE + ( RegNum / 8 );
                OcpRegNum = 0x10 + ( RegNum % 8 );
        } else {
                OcpPageNum = PageNum;
                OcpRegNum = RegNum;
        }

        OcpPageNum <<= 4;

        if ( OcpRegNum < 16 ) {
                OcpPhyAddress = 0;
        } else {
                OcpRegNum -= 16;
                OcpRegNum <<= 1;

                OcpPhyAddress = OcpPageNum + OcpRegNum;
        }


        return OcpPhyAddress;
}

u32 mdio_real_read_phy_ocp(struct rtl_hw *hw,
                           u32 RegAddr)
{
        u32 data32;
        int i, value = 0;

        data32 = RegAddr/2;
        data32 <<= OCPR_Addr_Reg_shift;

        RTL_W32(hw, PHYOCP, data32);
        for (i = 0; i < 100; i++) {
                udelay(1);

                if (RTL_R32(hw, PHYOCP) & OCPR_Flag)
                        break;
        }
        value = RTL_R32(hw, PHYOCP) & OCPDR_Data_Mask;

        return value;
}

u32 mdio_direct_read_phy_ocp(struct rtl_hw *hw,
                             u32 RegAddr)
{
        return mdio_real_read_phy_ocp(hw, RegAddr);
}

u32 rtl8125_mdio_read_phy_ocp(struct rtl_hw *hw,
                              u16 PageNum,
                              u32 RegAddr)
{
        u16 ocp_addr;

        ocp_addr = map_phy_ocp_addr(PageNum, RegAddr);

        return mdio_direct_read_phy_ocp(hw, ocp_addr);
}

u32 mdio_real_read(struct rtl_hw *hw,
                   u32 RegAddr)
{
        return rtl8125_mdio_read_phy_ocp(hw, hw->cur_page, RegAddr);
}

void mdio_real_write_phy_ocp(struct rtl_hw *hw,
                             u32 RegAddr,
                             u32 value)
{
        u32 data32;
        int i;

        data32 = RegAddr/2;
        data32 <<= OCPR_Addr_Reg_shift;
        data32 |= OCPR_Write | value;

        RTL_W32(hw, PHYOCP, data32);
        for (i = 0; i < 100; i++) {
                udelay(1);

                if (!(RTL_R32(hw, PHYOCP) & OCPR_Flag))
                        break;
        }
}

void mdio_direct_write_phy_ocp(struct rtl_hw *hw,
                               u32 RegAddr,
                               u32 value)
{
        mdio_real_write_phy_ocp(hw, RegAddr, value);
}

void rtl8125_mdio_write_phy_ocp(struct rtl_hw *hw,
                                u16 PageNum,
                                u32 RegAddr,
                                u32 value)
{
        u16 ocp_addr;

        ocp_addr = map_phy_ocp_addr(PageNum, RegAddr);

        mdio_direct_write_phy_ocp(hw, ocp_addr, value);
}

/*
void rtl8125_mdio_real_write_phy_ocp(struct rtl_hw *hw,
                u16 PageNum,
                u32 RegAddr,
                u32 value)
{
        u16 ocp_addr;

        ocp_addr = map_phy_ocp_addr(PageNum, RegAddr);

        mdio_direct_write_phy_ocp(hw, ocp_addr, value);
}
*/
void mdio_real_write(struct rtl_hw *hw, u32 RegAddr, u32 value)
{
        if (RegAddr == 0x1F) {
                hw->cur_page = value;
        }
        rtl8125_mdio_write_phy_ocp(hw, hw->cur_page, RegAddr, value);
}

u32 rtl8125_mdio_read(struct rtl_hw *hw, u32 RegAddr)
{
        //if (hw->rtk_enable_diag) return 0xffffffff;
        return mdio_real_read(hw, RegAddr);
}

void rtl8125_mdio_write(struct rtl_hw *hw, u32 RegAddr, u32 value)
{
        //if (hw->rtk_enable_diag) return;
        mdio_real_write(hw, RegAddr, value);
}
/*  no use
void ClearAndSetEthPhyBit(struct rtl_hw *hw, u8  addr, u16 clearmask, u16 setmask)
{
        u16 PhyRegValue;

        PhyRegValue = rtl8125_mdio_read(hw, addr);
        PhyRegValue &= ~clearmask;
        PhyRegValue |= setmask;
        rtl8125_mdio_write(hw, addr, PhyRegValue);
}

void rtl8125_clear_eth_phy_bit(struct rtl_hw *hw, u8 addr, u16 mask)
{
        ClearAndSetEthPhyBit(hw, addr, mask, 0);
}

void rtl8125_set_eth_phy_bit(struct rtl_hw *hw,  u8  addr, u16  mask)
{
        ClearAndSetEthPhyBit(hw, addr, 0, mask);
}
*/

void ClearAndSetEthPhyOcpBit(struct rtl_hw *hw, u16 addr, u16 clearmask, u16 setmask)
{
        u16 PhyRegValue;

        PhyRegValue = mdio_direct_read_phy_ocp(hw, addr);
        PhyRegValue &= ~clearmask;
        PhyRegValue |= setmask;
        mdio_direct_write_phy_ocp(hw, addr, PhyRegValue);
}

static int
rtl8125_check_hw_phy_mcu_code_ver(struct rtl_hw *hw)
{
        int ram_code_ver_match = 0;

        switch (hw->mcfg) {
        case CFG_METHOD_2 ... CFG_METHOD_5:
                mdio_direct_write_phy_ocp(hw, 0xA436, 0x801E);
                hw->hw_ram_code_ver = mdio_direct_read_phy_ocp(hw, 0xA438);
                break;
        default:
                hw->hw_ram_code_ver = ~0;
                break;
        }

        //printf("xing check print phy code hw version %d\n", hw->hw_ram_code_ver);

        if ( hw->hw_ram_code_ver == hw->sw_ram_code_ver) {
                ram_code_ver_match = 1;
                hw->HwHasWrRamCodeToMicroP = TRUE;
        }

        return ram_code_ver_match;
}

static void
rtl8125_write_hw_phy_mcu_code_ver(struct rtl_hw *hw)
{
        switch (hw->mcfg) {
        case CFG_METHOD_2 ... CFG_METHOD_5:
                mdio_direct_write_phy_ocp(hw, 0xA436, 0x801E);
                mdio_direct_write_phy_ocp(hw, 0xA438, hw->sw_ram_code_ver);
                hw->hw_ram_code_ver = hw->sw_ram_code_ver;
                break;
        }
}

bool
rtl8125_set_phy_mcu_patch_request(struct rtl_hw *hw)
{
        u16 gphy_val;
        u16 WaitCount;
        bool bSuccess = TRUE;

        SetEthPhyOcpBit(hw, 0xB820, BIT_4);

        WaitCount = 0;
        do {
                gphy_val = mdio_direct_read_phy_ocp(hw, 0xB800);
                udelay(100);
                WaitCount++;
        } while (!(gphy_val & BIT_6) && (WaitCount < 1000));

        if (!(gphy_val & BIT_6) && (WaitCount == 1000)) bSuccess = FALSE;

        if (!bSuccess)
                printf("rtl8125_set_phy_mcu_patch_request fail.\n");

        return bSuccess;

}


void ClearEthPhyOcpBit(struct rtl_hw *hw, u16 addr, u16 mask)
{
        ClearAndSetEthPhyOcpBit(hw,
                                addr,
                                mask,
                                0
                               );
}

bool
rtl8125_clear_phy_mcu_patch_request(struct rtl_hw *hw)
{
        u16 gphy_val;
        u16 WaitCount;
        bool bSuccess = TRUE;

        ClearEthPhyOcpBit(hw, 0xB820, BIT_4);

        WaitCount = 0;
        do {
                gphy_val = mdio_direct_read_phy_ocp(hw, 0xB800);
                udelay(100);
                WaitCount++;
        } while ((gphy_val & BIT_6) && (WaitCount < 1000));

        if ((gphy_val & BIT_6) && (WaitCount == 1000)) bSuccess = FALSE;

        if (!bSuccess)
                printf("rtl8125_clear_phy_mcu_patch_request fail.\n");

        return bSuccess;
}
//void rtl8125_enable_ocp_phy_power_saving(struct rtl_hw *hw)  no use

void
rtl8125_disable_ocp_phy_power_saving(struct rtl_hw *hw)
{
        u16 val;

        switch (hw->mcfg) {
        case CFG_METHOD_2 ... CFG_METHOD_5:
                val = mdio_direct_read_phy_ocp(hw, 0xC416);
                if (val != 0x0050) {
                        rtl8125_set_phy_mcu_patch_request(hw);
                        mdio_direct_write_phy_ocp(hw, 0xC416, 0x0000);
                        mdio_direct_write_phy_ocp(hw, 0xC416, 0x0500);
                        rtl8125_clear_phy_mcu_patch_request(hw);
                }
        }
}

static void
rtl8125_enable_phy_disable_mode(struct rtl_hw *hw)
{

        switch (hw->HwSuppCheckPhyDisableModeVer) {
        case 3:
                RTL_W8(hw, 0xF2, RTL_R8(hw, 0xF2) | BIT_5);
                break;
        }

}

static void
rtl8125_disable_phy_disable_mode(struct rtl_hw *hw)
{

        switch (hw->HwSuppCheckPhyDisableModeVer) {
        case 3:
                RTL_W8(hw, 0xF2, RTL_R8(hw, 0xF2) & ~BIT_5);
                break;
        }

        mdelay(1);

}

static u8
rtl8125_is_phy_disable_mode_enabled(struct rtl_hw *hw)
{
        u8 phy_disable_mode_enabled = FALSE;

        switch (hw->HwSuppCheckPhyDisableModeVer) {
        case 3:
                if (RTL_R8(hw, 0xF2) & BIT_5)
                        phy_disable_mode_enabled = TRUE;
                break;
        }

        return phy_disable_mode_enabled;
}

static u8
rtl8125_is_gpio_low(struct rtl_hw *hw)
{
        u8 gpio_low = FALSE;

        switch (hw->HwSuppCheckPhyDisableModeVer) {
        case 3:
                if (!(rtl8125_mac_ocp_read(hw, 0xDC04) & BIT_13))
                        gpio_low = TRUE;
                break;
        }

        return gpio_low;
}

static u8
rtl8125_is_in_phy_disable_mode(struct rtl_hw *hw)
{
        u8 in_phy_disable_mode = FALSE;

        if (rtl8125_is_phy_disable_mode_enabled(hw) && rtl8125_is_gpio_low(hw))
                in_phy_disable_mode = TRUE;

        return in_phy_disable_mode;
}

int
rtl8125_init_hw_phy_mcu(struct rtl_hw *hw)
{
        u8 require_disable_phy_disable_mode = FALSE;

        if (hw->HwIcVerUnknown) {
                printf("xing get hw version unknown!\n");
                return FALSE;
        }

        if (rtl8125_check_hw_phy_mcu_code_ver(hw)) {
                //printf("xing check phy mcu code version sucess!\n");
                //we need not update phy fw
                return FALSE;
        }

        if ((hw->HwSuppCheckPhyDisableModeVer > 0) && rtl8125_is_in_phy_disable_mode(hw))
                require_disable_phy_disable_mode = TRUE;

        if (require_disable_phy_disable_mode)
                rtl8125_disable_phy_disable_mode(hw);

        hw->hw_ops.hw_phy_mcu_config(hw);

        if (require_disable_phy_disable_mode)
                rtl8125_enable_phy_disable_mode(hw);

        rtl8125_write_hw_phy_mcu_code_ver(hw);

        rtl8125_mdio_write(hw,0x1F, 0x0000);

        hw->HwHasWrRamCodeToMicroP = TRUE;

        return TRUE;
}

static void
rtl8125_xmii_reset_enable(struct rtl_hw *hw)
{
        int i, val = 0;

        if (rtl8125_is_in_phy_disable_mode(hw))
                return;

        rtl8125_mdio_write(hw, 0x1f, 0x0000);
        rtl8125_mdio_write(hw, MII_ADVERTISE, rtl8125_mdio_read(hw, MII_ADVERTISE) &
                           ~(ADVERTISE_10HALF | ADVERTISE_10FULL |
                             ADVERTISE_100HALF | ADVERTISE_100FULL));
        rtl8125_mdio_write(hw, MII_CTRL1000, rtl8125_mdio_read(hw, MII_CTRL1000) &
                           ~(ADVERTISE_1000HALF | ADVERTISE_1000FULL));
        mdio_direct_write_phy_ocp(hw, 0xA5D4, mdio_direct_read_phy_ocp(hw, 0xA5D4) & ~(RTK_ADVERTISE_2500FULL));
        rtl8125_mdio_write(hw, MII_BMCR, BMCR_RESET | BMCR_ANENABLE);

        for (i = 0; i < 2500; i++) {
                val = rtl8125_mdio_read(hw, MII_BMCR) & BMCR_RESET;

                if (!val) {
                        return;
                }

                mdelay(1);
        }
}

static void
rtl8125_phy_restart_nway(struct rtl_hw *hw)
{

        if (rtl8125_is_in_phy_disable_mode(hw)) return;

        rtl8125_mdio_write(hw, 0x1F, 0x0000);
        rtl8125_mdio_write(hw, MII_BMCR, BMCR_RESET | BMCR_ANENABLE | BMCR_ANRESTART);
}

static int
rtl8125_phy_setup_force_mode(struct rtl_hw *hw, u32 speed, u8 duplex)
{
        int ret = FALSE;
        u16 bmcr_true_force = 0;

        if (rtl8125_is_in_phy_disable_mode(hw)) return ret;

        if ((speed == SPEED_10) && (duplex == DUPLEX_HALF)) {
                bmcr_true_force = BMCR_SPEED10;
        } else if ((speed == SPEED_10) && (duplex == DUPLEX_FULL)) {
                bmcr_true_force = BMCR_SPEED10 | BMCR_FULLDPLX;
        } else if ((speed == SPEED_100) && (duplex == DUPLEX_HALF)) {
                bmcr_true_force = BMCR_SPEED100;
        } else if ((speed == SPEED_100) && (duplex == DUPLEX_FULL)) {
                bmcr_true_force = BMCR_SPEED100 | BMCR_FULLDPLX;
        } else if ((speed == SPEED_1000) && (duplex == DUPLEX_FULL) &&
                   hw->HwSuppGigaForceMode) {
                bmcr_true_force = BMCR_SPEED1000 | BMCR_FULLDPLX;
        } else {
                //log print
                return FALSE;
        }

        rtl8125_mdio_write(hw, 0x1F, 0x0000);
        rtl8125_mdio_write(hw, MII_BMCR, bmcr_true_force);

        return TRUE;
}

void
rtl8125_link_option(u8 *aut, u32 *spd, u8 *dup, u32 *adv)
{
        if ((*spd != SPEED_2500) && (*spd != SPEED_1000) &&
            (*spd != SPEED_100) && (*spd != SPEED_10))
                *spd = SPEED_2500;

        if ((*dup != DUPLEX_FULL) && (*dup != DUPLEX_HALF))
                *dup = DUPLEX_FULL;

        if ((*aut != AUTONEG_ENABLE) && (*aut != AUTONEG_DISABLE))
                *aut = AUTONEG_ENABLE;

        *adv &= (ADVERTISE_10_HALF |
                 ADVERTISE_10_FULL |
                 ADVERTISE_100_HALF |
                 ADVERTISE_100_FULL |
                 ADVERTISE_1000_HALF |
                 ADVERTISE_1000_FULL |
                 ADVERTISE_2500_FULL);//tocheck0x8000 or 0x80
        if (*adv == 0)
                *adv = (ADVERTISE_10_HALF |
                        ADVERTISE_10_FULL |
                        ADVERTISE_100_HALF |
                        ADVERTISE_100_FULL |
                        ADVERTISE_1000_HALF |
                        ADVERTISE_1000_FULL |
                        ADVERTISE_2500_FULL);
}

static int
rtl8125_set_speed_xmii(struct rtl_hw *hw, u8 autoneg, u32 speed, u8 duplex, u32 adv)
{
        int auto_nego = 0;
        int giga_ctrl = 0;
        int ctrl_2500 = 0;
        int rc = -EINVAL;

        //Disable Giga Lite
        ClearEthPhyOcpBit(hw, 0xA428, BIT_9);
        ClearEthPhyOcpBit(hw, 0xA5EA, BIT_0);

        if (speed != SPEED_2500 &&
            (speed != SPEED_1000) &&
            (speed != SPEED_100) &&
            (speed != SPEED_10)) {
                speed = SPEED_2500;
                duplex = DUPLEX_FULL;
        }

        giga_ctrl = rtl8125_mdio_read(hw, MII_CTRL1000);
        giga_ctrl &= ~(ADVERTISE_1000HALF | ADVERTISE_1000FULL);
        ctrl_2500 = mdio_direct_read_phy_ocp(hw, 0xA5D4);
        ctrl_2500 &= ~(RTK_ADVERTISE_2500FULL);

        if (autoneg == AUTONEG_ENABLE) {
                /*n-way force*/
                auto_nego = rtl8125_mdio_read(hw, MII_ADVERTISE);
                auto_nego &= ~(ADVERTISE_10HALF | ADVERTISE_10FULL |
                               ADVERTISE_100HALF | ADVERTISE_100FULL |
                               ADVERTISE_PAUSE_CAP | ADVERTISE_PAUSE_ASYM);

                if (adv & ADVERTISE_10_HALF)
                        auto_nego |= ADVERTISE_10HALF;
                if (adv & ADVERTISE_10_FULL)
                        auto_nego |= ADVERTISE_10FULL;
                if (adv & ADVERTISE_100_HALF)
                        auto_nego |= ADVERTISE_100HALF;
                if (adv & ADVERTISE_100_FULL)
                        auto_nego |= ADVERTISE_100FULL;
                if (adv & ADVERTISE_1000_HALF)
                        giga_ctrl |= ADVERTISE_1000HALF;
                if (adv & ADVERTISE_1000_FULL)
                        giga_ctrl |= ADVERTISE_1000FULL;
                if (adv & ADVERTISE_2500_FULL)//tocheck
                        ctrl_2500 |= RTK_ADVERTISE_2500FULL;

                //flow control
                if (hw->mtu <= ETH_DATA_LEN)
                        auto_nego |= ADVERTISE_PAUSE_CAP|ADVERTISE_PAUSE_ASYM;

                hw->phy_auto_nego_reg = auto_nego;
                hw->phy_1000_ctrl_reg = giga_ctrl;

                hw->phy_2500_ctrl_reg = ctrl_2500;//tocheck

                rtl8125_mdio_write(hw, 0x1f, 0x0000);
                rtl8125_mdio_write(hw, MII_ADVERTISE, auto_nego);
                rtl8125_mdio_write(hw, MII_CTRL1000, giga_ctrl);
                mdio_direct_write_phy_ocp(hw, 0xA5D4, ctrl_2500);
                rtl8125_phy_restart_nway(hw);
                mdelay(20);
        } else {
                /*true force*/
                if (speed == SPEED_10 || speed == SPEED_100 ||
                    (speed == SPEED_1000 && duplex == DUPLEX_FULL &&
                     hw->HwSuppGigaForceMode)) {
                        rtl8125_phy_setup_force_mode(hw, speed, duplex);
                } else
                        goto out;
        }
        hw->autoneg = autoneg;
        hw->speed = speed;
        hw->duplex = duplex;
        hw->advertising = adv;

        rc = 0;
out:
        return rc;

}

static void
rtl8125_wait_phy_ups_resume(struct rtl_hw *hw, u16 PhyState)
{
        u16 TmpPhyState;
        int i=0;

        switch (hw->mcfg) {
        case CFG_METHOD_2 ... CFG_METHOD_5:
                do {
                        TmpPhyState = mdio_direct_read_phy_ocp(hw, 0xA420);
                        TmpPhyState &= 0x7;
                        mdelay(1);
                        i++;
                } while ((i < 100) && (TmpPhyState != PhyState));
        }
}

static void
rtl8125_phy_power_up(struct rtl_hw *hw)
{

        if (rtl8125_is_in_phy_disable_mode(hw))
                return;

        rtl8125_mdio_write(hw, 0x1F, 0x0000);

        rtl8125_mdio_write(hw, MII_BMCR, BMCR_ANENABLE);//tocheck MII_BMCR BMCR_ANENABLE

        //wait ups resume (phy state 3)
        switch (hw->mcfg) {
        case CFG_METHOD_2 ... CFG_METHOD_5:
                rtl8125_wait_phy_ups_resume(hw, 3);
        }
}

void rtl8125_powerup_pll(struct rtl_hw *hw)
{
        switch (hw->mcfg) {
        case CFG_METHOD_2 ... CFG_METHOD_5:
                RTL_W8(hw, PMCH, RTL_R8(hw, PMCH) | BIT_7 | BIT_6);
        }

        rtl8125_phy_power_up(hw);
}

static void
rtl8125_phy_power_down(struct rtl_hw *hw)
{
        rtl8125_mdio_write(hw, 0x1F, 0x0000);
        rtl8125_mdio_write(hw, MII_BMCR, BMCR_ANENABLE | BMCR_PDOWN);
}

void rtl8125_powerdown_pll(struct rtl_hw *hw)
{

        rtl8125_phy_power_down(hw);

        switch (hw->mcfg) {
        case CFG_METHOD_2 ... CFG_METHOD_5:
                RTL_W8(hw, PMCH, RTL_R8(hw, PMCH) & ~BIT_7);
                break;
        }

        //aldps pwrsave settings
        //switch (hw->mcfg) {
        //case CFG_METHOD_14 ... CFG_METHOD_15:
        //        RTL_W8(hw, 0xD0, RTL_R8(hw, 0xD0) & ~BIT_6);
        //        break;
        //case CFG_METHOD_16 ... CFG_METHOD_33:
        //        RTL_W8(hw, 0xD0, RTL_R8(hw, 0xD0) & ~BIT_6);
        //        RTL_W8(hw, 0xF2, RTL_R8(hw, 0xF2) & ~BIT_6);
        //        break;
        //}
}


void SetEthPhyOcpBit(struct rtl_hw *hw,  u16 addr, u16 mask)
{
        ClearAndSetEthPhyOcpBit(hw,
                                addr,
                                0,
                                mask
                               );
}

int
rtl8125_set_speed(struct rtl_hw *hw)
{
        int rc = TRUE;
        if (FALSE == rtl8125_set_speed_xmii(hw, hw->autoneg, hw->speed, hw->duplex, hw->advertising)) {
                //log set speed fail;
                rc = FALSE;
        }
        return rc;
}

static void
rtl8125_disable_aldps(struct rtl_hw *hw)
{
        u16 TmpUshort;

        TmpUshort = mdio_real_read_phy_ocp(hw, 0xA430);
        if (TmpUshort & BIT_2) {
                u32 Timeout = 0, WaitCount = 200;

                ClearEthPhyOcpBit(hw, 0xA430, BIT_2);

                do {
                        udelay(100);

                        TmpUshort = rtl8125_mac_ocp_read(hw, 0xE908);

                        Timeout++;
                } while (!(TmpUshort & BIT_7) && Timeout < WaitCount);
        }
}

static void rtl8125_disable_eee(struct rtl_hw *hw)
{
        switch (hw->mcfg) {
        case CFG_METHOD_2:
        case CFG_METHOD_3:
                ClearMcuAccessRegBit(hw, 0xE040, (BIT_1 | BIT_0));
                ClearMcuAccessRegBit(hw, 0xEB62, (BIT_2 | BIT_1));

                ClearEthPhyOcpBit(hw, 0xA432, BIT_4);
                ClearEthPhyOcpBit(hw, 0xA5D0, (BIT_2 | BIT_1));
                ClearEthPhyOcpBit(hw, 0xA6D4, BIT_0);

                ClearEthPhyOcpBit(hw, 0xA6D8, BIT_4);
                ClearEthPhyOcpBit(hw, 0xA428, BIT_7);
                ClearEthPhyOcpBit(hw, 0xA4A2, BIT_9);
                break;
        case CFG_METHOD_4:
        case CFG_METHOD_5:
                ClearMcuAccessRegBit(hw, 0xE040, (BIT_1 | BIT_0));

                ClearEthPhyOcpBit(hw, 0xA5D0, (BIT_2 | BIT_1));
                ClearEthPhyOcpBit(hw, 0xA6D4, BIT_0);

                ClearEthPhyOcpBit(hw, 0xA6D8, BIT_4);
                ClearEthPhyOcpBit(hw, 0xA428, BIT_7);
                ClearEthPhyOcpBit(hw, 0xA4A2, BIT_9);
                break;
        default:
                // Not Support EEE
                break;
        }

        /*Advanced EEE*/
        switch (hw->mcfg) {
        case CFG_METHOD_2:
        case CFG_METHOD_3:
        case CFG_METHOD_4:
        case CFG_METHOD_5:
                rtl8125_set_phy_mcu_patch_request(hw);
                ClearMcuAccessRegBit(hw, 0xE052, BIT_0);
                ClearEthPhyOcpBit(hw, 0xA442, BIT_12 | BIT_13);
                ClearEthPhyOcpBit(hw, 0xA430, BIT_15);
                rtl8125_clear_phy_mcu_patch_request(hw);
                break;
        }
}

void
rtl8125_hw_phy_config(struct rtl_hw *hw)
{
        rtl8125_xmii_reset_enable(hw);

        rtl8125_set_hw_phy_before_init_phy_mcu(hw);

        if (FALSE == rtl8125_init_hw_phy_mcu(hw)) {
                //TODO log print
                //printf("xing bypass phy mcu init!\n");
        }

        hw->hw_ops.hw_phy_config(hw);

        switch (hw->mcfg) {
        case CFG_METHOD_2 ... CFG_METHOD_5:
                rtl8125_disable_aldps(hw);
                break;
        }

        rtl8125_mdio_write(hw, 0x1F, 0x0000);

        if (HW_HAS_WRITE_PHY_MCU_RAM_CODE(hw)) {
                //if (hw->eee_enabled)
                //        rtl8125_enable_eee(hw);
                //else
                rtl8125_disable_eee(hw);
        }
}
/*
void
rtl8125_enable_phy_aldps(struct rtl_hw *hw)
{
        //enable aldps
        //GPHY OCP 0xA430 bit[2] = 0x1 (en_aldps)
        SetEthPhyOcpBit(hw, 0xA430, BIT_2);
}
*/
//tocheck howard
void
rtl8125_set_hw_phy_before_init_phy_mcu(struct rtl_hw *hw)
{
        u16 PhyRegValue;

        switch (hw->mcfg) {
        case CFG_METHOD_4:
                mdio_direct_write_phy_ocp(hw, 0xBF86, 0x9000);

                SetEthPhyOcpBit(hw, 0xC402, BIT_10);
                ClearEthPhyOcpBit(hw, 0xC402, BIT_10);

                PhyRegValue = mdio_direct_read_phy_ocp(hw, 0xBF86);
                PhyRegValue &= (BIT_1 | BIT_0);
                if (PhyRegValue != 0)
                        printf("PHY watch dog not clear, value = 0x%x \n", PhyRegValue);

                mdio_direct_write_phy_ocp(hw, 0xBD86, 0x1010);
                mdio_direct_write_phy_ocp(hw, 0xBD88, 0x1010);

                ClearAndSetEthPhyOcpBit(hw,
                                        0xBD4E,
                                        BIT_11 | BIT_10,
                                        BIT_11);
                ClearAndSetEthPhyOcpBit(hw,
                                        0xBF46,
                                        BIT_11 | BIT_10 | BIT_9 | BIT_8,
                                        BIT_10 | BIT_9 | BIT_8);
                break;
        }
}
